from ._api_response_object import object_helper
from .workspace import Workspace
from .ssh_key import SSHKey


class Organization(object):
    """ Organization class instaniated with api response attributes for organizations. """

    def __init__(self, organization, api_handler):
        self._api_handler = api_handler

        organization = organization.attributes
        self.name = organization.name
        self.external_id = organization.external_id
        self.created_at = organization.created_at
        self.email = organization.email
        self.session_timeout = organization.session_timeout
        self.session_remember = organization.session_remember
        self.collaborator_auth_policy = organization.collaborator_auth_policy
        self.enterprise_plan = organization.enterprise_plan
        self.plan_expired = organization.plan_expired
        self.cost_estimation_enabled = organization.cost_estimation_enabled
        self.permissions = organization.permissions
        self.fair_run_queuing_enabled = organization.fair_run_queuing_enabled
        self.saml_enabled = organization.saml_enabled
        self.two_factor_conformant = organization.two_factor_conformant
        self.preview_request = organization.preview_request

    def __str__(self):
        return str(self.__dict__)

    def list_workspaces(self):
        """
        Returns list of all workspace objects in the provided organization.
        If data response is empty (empty page) stop making queries.
        """
        workspaces = []
        params = {"page[size]": 100, "page[number]": 1}
        while True:
            response = self._api_handler.call(
                uri=f'organizations/{self.name}/workspaces', params=params)
            if not response.data:
                break
            for workspace in response.data:
                workspaces.append(
                    Workspace(workspace=object_helper(workspace),
                              organization_name=self.name,
                              api_handler=self._api_handler))
            params["page[number]"] += 1
        return workspaces

    def search_workspaces(self, keyword):
        """ Get workspace objects by substring of name. """
        workspaces = []
        for workspace in self.list_workspaces():
            if keyword.lower() in workspace.name.lower():
                workspaces.append(workspace)
        return workspaces

    def create_workspace(self,
                         name,
                         tf_version=None,
                         working_directory="",
                         auto_apply=False,
                         queue_all_runs=False,
                         vcs_repo=None,
                         trigger_prefixes=[]):
        """
        Creates workspace, must pass vcs payload as dictionary. Returns response json. Example vcs-repo payload:
        {
            "identifier": "skierkowski/terraform-test-proj",
            "oauth-token-id": "ot-hmAyP66qk2AMVdbJ",
            "branch": "",
            "default-branch": True
        }

        """
        payload = {
            "data": {
                "attributes": {
                    "name": name,
                    "terraform_version": tf_version,
                    "working-directory": working_directory,
                    "auto-apply": auto_apply,
                    "queue-all-runs": queue_all_runs,
                    "trigger-prefixes": trigger_prefixes,
                    "vcs-repo": vcs_repo
                },
                "type": "workspaces"
            }
        }
        return self._api_handler.call(
            uri=f'organizations/{self.name}/workspaces',
            method='post',
            json=payload).data

    def delete_workspace(self, name):
        """ Delete a workspace by name within the organization. """
        return self._api_handler.call(
            method='delete', uri=f'organizations/{self.name}/workspaces/{name}')

    def get_workspace(self, name):
        """ Referred to as show workspace in official documentation, returns a single workspace object. """
        response = self._api_handler.call(
            uri=f'organizations/{self.name}/workspaces/{name}')
        return Workspace(workspace=object_helper(response.data),
                         organization_name=self.name,
                         api_handler=self._api_handler)

    def list_ssh_keys(self):
        """ List all SSH Keys in a given account. """
        ssh_keys = []
        response = self._api_handler.call(
            uri=f'organizations/{self.name}/ssh-keys')
        for ssh_key in response.data:
            ssh_keys.append(SSHKey(object_helper(ssh_key), self._api_handler))
        return ssh_keys

    def create_ssh_key(self, name, key):
        """ Create an ssh key, requires the string of the private key. """
        payload = {
            "data": {
                "type": "ssh-keys",
                "attributes": {
                    "name": name,
                    "value": key
                }
            }
        }
        response = self._api_handler.call(
            method='post',
            uri=f'organizations/{self.name}/ssh-keys',
            json=payload).data
        return SSHKey(ssh_key=object_helper(response),
                      api_handler=self._api_handler)



import collections
import requests
from .exceptions import APIException


class APICaller(object):
    def __init__(self,base_url, headers):
        self._base_url = base_url
        self._headers = headers

    def call(self, uri, method="get", *args, **kwargs):
        requester = getattr(requests, method.lower())
        url = self._base_url + uri
        response = requester(url=url, headers=self._headers, *args, **kwargs)
        if response.status_code < 400:
            if method in ["get", "post", "patch", "put"]:
                response_json = response.json()
                if response_json and "data" in response_json:
                    return APIResponse(response_json)
                else:
                    return response.status_code
            elif method == "delete":
                return response.status_code
        else:
            raise APIException(f"Error: {response.status_code}: {response.content}, {url}", response)


class APIResponse(object):
    def __init__(self, response):
        self.data = response.get("data", [])
        self.meta = response.get("meta", [])
        self.links = response.get("links", [])
        self.included = response.get("included", [])
        self.errors = response.get("errors", [])

    def __str__(self):
        if self.data:
            if isinstance(self.data, collections.abc.Iterable):
                return str(self.data)



import requests

class Run(object):

    def __init__(self, run, api_handler):
        self.id = run.id
        self._api_handler = api_handler

        relationships = run.relationships
        self.workspace_id = relationships.workspace.data.id
        self.apply_id = relationships.apply.data.id
        self.configuration_version = relationships.configuration_version.data.id
        self.plan_id = relationships.plan.data.id
        self.run_events = relationships.run_events.data
        self.policy_checks = relationships.policy_checks.data
        self.comments = relationships.comments.data

        attributes = run.attributes
        self.actions = attributes.actions
        self.canceled_at = attributes.canceled_at
        self.created_at = attributes.created_at
        self.has_changes = attributes.has_changes
        self.is_destroy = attributes.is_destroy
        self.message = attributes.message
        self.plan_only = attributes.plan_only
        self.source = attributes.source
        self.status_timestamps = attributes.status_timestamps
        self.status = attributes.status
        self.trigger_reason = attributes.trigger_reason
        self.permissions = attributes.permissions

    def __str__(self):
        return str(self.__dict__)

    def apply(self, comment=None):
        """ Apply a run on instantiate workspace, returns response code. """
        return self._api_handler.call(method='post',
                                      uri=f'runs/{self.id}/actions/apply',
                                      json={'comment': comment})

    def discard(self, comment=None):
        """ Discard a run on instantiate workspace, returns response code. """
        return self._api_handler.call(method='post',
                                      uri=f'runs/{self.id}/actions/discard',
                                      json={'comment': comment})

    def cancel(self, comment=None):
        """ Cancel a run on instantiate workspace, returns response code. """
        return self._api_handler.call(method='post',
                                      uri=f'runs/{self.id}/actions/cancel',
                                      json={'comment': comment})

    def force_cancel(self, comment=None):
        return self._api_handler.call(
            method='post',
            uri=f'runs/{self.id}/actions/force-cancel',
            json={'comment': comment})

    def force_execute(self, comment=None):
        return self._api_handler.call(
            method='post',
            uri=f'runs/{self.id}/actions/force-execute',
            json={'comment': comment})

    def get_plan(self):
        return self._api_handler.call(uri=f'plans/{self.plan_id}').data

    def get_plan_output(self):
        """ Return plan output string of run. """
        plan = self.get_plan()
        return requests.get(plan['attributes']['log-read-url']).content



class Variable(object):

    def __init__(self, variable, api_handler):
        self._api_handler = api_handler

        self.id = variable.id
        attributes = variable.attributes
        self.key = attributes.key
        self.value = attributes.value
        self.category = attributes.category
        self.hcl = attributes.hcl
        self.sensitive = attributes.sensitive
        self.workspace_id = variable.relationships.configurable.data.id

    def __str__(self):
        return str(self.__dict__)

    def update(self,
               key=None,
               value=None,
               category=None,
               hcl=None,
               sensitive=None,
               description=None):
        """ Update variable with supplied fields. """
        defaults = {
            'key': self.key,
            'value': self.value,
            'category': self.category,
            'hcl': self.hcl,
            'sensitive': self.sensitive,
            'description': ""
        }

        arguments = locals()
        del arguments['self']

        for arg, value in arguments.items():
            if value != None:
                defaults[arg] = value

        payload = {
            "data": {
                "type": "vars",
                "attributes": {
                    "key": defaults["key"],
                    "value": defaults["value"],
                    "description": defaults["description"],
                    "category": defaults["category"],
                    "hcl": defaults["hcl"],
                    "sensitive": defaults["sensitive"]
                },
                "relationships": {
                    "workspace": {
                        "data": {
                            "id": self.workspace_id,
                            "type": "workspaces"
                        }
                    }
                }
            }
        }
        return self._api_handler.call(uri=f'vars/{self.id}',
                                      method='patch',
                                      json=payload).data

    def delete(self):
        return self._api_handler.call(uri=f'vars/{self.id}', method='delete')



import time
from .exceptions import APIException
from ._api_response_object import object_helper, collections
from .run import Run
from .variable import Variable


class Workspace(object):

    def __init__(self, workspace, organization_name, api_handler):
        self._api_handler = api_handler
        self.id = workspace.id
        self.organization_name = organization_name

        relationships = workspace.relationships
        self.current_run = relationships.current_run
        self.latest_run = relationships.latest_run
        self.current_state_version = relationships.current_state_version

        attributes = workspace.attributes
        self.name = attributes.name
        self.auto_apply = attributes.auto_apply
        self.created_at = attributes.created_at
        self.environment = attributes.environment
        self.locked = attributes.locked
        self.queue_all_runs = attributes.queue_all_runs
        self.terraform_version = attributes.terraform_version
        self.working_directory = attributes.working_directory
        self.speculative_enabled = attributes.speculative_enabled
        self.latest_change_at = attributes.latest_change_at
        self.operations = attributes.operations
        self.vcs_repo = attributes.vcs_repo
        self.permissions = attributes.permissions
        self.actions = attributes.actions
        self.description = attributes.description
        self.file_triggers_enabled = attributes.file_triggers_enabled
        self.trigger_prefixes = attributes.trigger_prefixes
        self.source = attributes.source
        self.source_name = attributes.source_name
        self.source_url = attributes.source_url
        self.links = workspace.links

    def __str__(self):
        return str(self.__dict__)

    def run(self, message='Ran from Terraform API', destroy_flag=False):
        """ Applies a run in an instantiated workspace. Can optionally destroy. """
        payload = {
            "data": {
                "attributes": {
                    "is-destroy": destroy_flag,
                    "message": message
                },
                "type": "runs",
                "relationships": {
                    "workspace": {
                        "data": {
                            "type": "workspaces",
                            "id": self.id
                        }
                    }
                }
            }
        }
        response = self._api_handler.call(method='post',
                                          uri='runs',
                                          json=payload).data
        return Run(object_helper(response), self._api_handler)

    def plan_apply(self, message, destroy_flag=False, timeout=120):
        """ Prints Plan and Apply output of Workspace, if stuck in queue will eventually timeout. """
        counter = 0
        run = self.run(message, destroy_flag)
        while run.status == 'pending':
            if counter > 15:
                print('Plan queue timeout.')
                return
            run = self.get_run(run.id)
            counter += 5
            time.sleep(5)

        counter = 0
        while run.status != 'planned' and counter < timeout:
            time.sleep(10)
            counter += 10
            run = self.get_run(run_id=run.id)
            if run.status == 'planned_and_finished':
                print('No changes detected.')
                return
        print(run.get_plan_output())

        print(f'Applying run: {self.id}: {run.id}')
        run.apply()
        run = self.get_run(run.id)
        while run.status == 'applying':
            time.sleep(5)
            run = self.get_run(run_id=run.id)
        print('Run Completed.')
        return

    def delete(self):
        """ Remove instantiated workspace. Returns response json. """
        return self._api_handler.call(
            method='delete',
            uri=f'organizations/{self.organization_name}/workspaces/{self.name}'
        )

    def update(self,
               name=None,
               tf_version=None,
               working_directory=None,
               auto_apply=None,
               queue_all_runs=None,
               vcs_repo=None,
               trigger_prefixes=None):
        """ Updates workspace. Uses instantiated defaults to populate payload with non-supplied values. """
        defaults = {
            'name': self.name,
            'tf_version': self.terraform_version,
            'working_directory': self.working_directory,
            'auto_apply': self.auto_apply,
            'queue_all_runs': self.queue_all_runs,
            'vcs_repo': self.vcs_repo,
            'trigger_prefixes': self.trigger_prefixes
        }

        arguments = locals()
        del arguments['self']

        for arg, value in arguments.items():
            if value != None:
                defaults[arg] = value
        payload = {
            "data": {
                "attributes": {
                    "name": defaults['name'],
                    "terraform_version": defaults['tf_version'],
                    "working-directory": defaults['working_directory'],
                    "auto-apply": defaults['auto_apply'],
                    "queue-all-runs": defaults['queue_all_runs'],
                    "trigger-prefixes": defaults['trigger_prefixes'],
                    "vcs-repo": defaults['vcs_repo']
                },
                "type": "workspaces"
            }
        }
        return self._api_handler.call(
            uri=f'organizations/{self.organization_name}/workspaces/{self.name}',
            method='patch',
            json=payload).data

    def lock(self, reason=""):
        """ Lock workspace. Returns reponse json. """
        payload = {"reason": reason}
        return self._api_handler.call(uri=f'workspaces/{self.id}/actions/lock',
                                      method='post',
                                      json=payload).data

    def unlock(self, reason=""):
        """ Unlocks workspace. Returns response json. """
        payload = {"reason": reason}
        return self._api_handler.call(
            uri=f'workspaces/{self.id}/actions/unlock',
            method='post',
            json=payload).data

    def force_unlock(self, reason):
        payload = {"reason": reason}
        return \
            self._api_handler.call(uri=f'workspaces/{self.id}/actions/force-unlock', method='post', json=payload).data

    def assign_ssh_key(self, ssh_key_id=''):
        """ Assigns ssh key requires ssh-key id found from the ssh-key endpoint. Returns response json. """
        payload = {
            "data": {
                "attributes": {
                    "id": ssh_key_id
                },
                "type": "workspaces"
            }
        }
        return \
            self._api_handler.call(uri=f'workspaces/{self.id}/relationships/ssh-key', method='patch', json=payload).data

    def unassign_ssh_key(self):
        """ Unassigns the ssh key currently assigned to the workspace. """
        payload = {"data": {"attributes": {"id": None}, "type": "workspaces"}}
        return \
            self._api_handler.call(uri=f'workspaces/{self.id}/relationships/ssh-key', method='patch', json=payload).data

    def get_current_state_version(self):
        """ Gets current statefile for a workspace, returns None if none found. """
        try:
            return object_helper(
                self._api_handler.call(
                    uri=f'workspaces/{self.id}/current-state-version').data)
        except APIException:
            return None

    # Create a class for handling variables with an update and delete method included instead of using workspace class.
    def list_variables(self):
        """ Returns list of variable objects for workspace for modification. """
        variables = []
        params = {
            "filter[workspace][name]": self.name,
            "filter[organization][name]": self.organization_name
        }
        response = self._api_handler.call(uri=f'vars', params=params).data
        for variable in response:
            variables.append(
                Variable(object_helper(variable), self._api_handler))
        return variables

    def create_variable(self,
                        key,
                        value,
                        category="env",
                        hcl=False,
                        sensitive=False,
                        description=""):
        """ Create a new variable in a workspace. """
        payload = {
            "data": {
                "type": "vars",
                "attributes": {
                    "key": key,
                    "value": value,
                    "description": description,
                    "category": category,
                    "hcl": hcl,
                    "sensitive": sensitive
                },
                "relationships": {
                    "workspace": {
                        "data": {
                            "id": self.id,
                            "type": "workspaces"
                        }
                    }
                }
            }
        }
        return self._api_handler.call(uri=f'vars', method='post',
                                      json=payload).data

    def get_run(self, run_id):
        run = self._api_handler.call(uri=f'/runs/{run_id}').data
        return Run(run=object_helper(run), api_handler=self._api_handler)

    def list_runs(self):
        runs = []
        response = self._api_handler.call(uri=f'workspaces/{self.id}/runs').data
        for run in response:
            runs.append(
                Run(run=object_helper(run), api_handler=self._api_handler))
        return runs



class SSHKey(object):

    def __init__(self, ssh_key, api_handler):
        self._api_handler = api_handler
        self.id = ssh_key.id
        self.name = ssh_key.attributes.name

    def __str__(self):
        return str(self.__dict__)

    def update(self, name):
        payload = {"data": {"attributes": {"name": name}}}
        return self._api_handler.call(method='patch',
                                      uri=f'ssh-keys/{self.id}',
                                      json=payload)

    def delete(self):
        return self._api_handler.call(method='delete',
                                      uri=f'ssh-keys/{self.id}')



from .handler import APICaller
from ._api_response_object import object_helper
from .organization import Organization


class Client():
    """
    Base client class than can modify organizations and perform other administrative tasks, for workspace or granular changes
    organization helper class needs to be instiated. Set token and V2 API URL using the non default 'init' constructor method.
    """

    def __init__(self):
        return

    def init(self, token, url, ssl_verification=True, version='v2'):
        self.token = token
        self.url = url + f'/api/{version}/'
        self.headers = {
            'Content-Type': 'application/vnd.api+json',
            'Authorization': f'Bearer {token}'
        }
        self.ssl_verification = ssl_verification
        self._api_handler = APICaller(base_url=self.url, headers=self.headers)

    def list_organizations(self):
        """ Returns list of instantiated class objects for workspaces. """
        organizations = []
        response = self._api_handler.call(uri='organizations')
        for organization in response.data:
            organizations.append(
                Organization(organization=object_helper(organization),
                             api_handler=self._api_handler))
        return organizations

    def set_organization(self, id):
        """ Instantiates a given organization class based off the passed id/name. """
        response = self._api_handler.call(uri=f'organizations/{id}')
        return Organization(organization=object_helper(response.data),
                            api_handler=self._api_handler)

    def create_organization(self,
                            name,
                            email,
                            session_timeout=20160,
                            session_remember=20160,
                            collaborator_auth_policy='password',
                            cost_estimation_enabled=False,
                            owners_team_saml_role_id=''):
        payload = {
            "data": {
                "type": "organizations",
                "attributes": {
                    "name": name,
                    "email": email,
                    "session-timeout": session_timeout,
                    "session-remember": session_remember,
                    "collaborator-auth-policy": collaborator_auth_policy,
                    "owners-team-saml-role-id": owners_team_saml_role_id,
                    "cost-estimation-enabled": cost_estimation_enabled
                }
            }
        }
        response = self._api_handler.call(method='post',
                                          uri=f'organizations',
                                          json=payload)
        return response.data

    def destroy_organization(self, name):
        return self._api_handler.call(method='delete',
                                      uri=f"organizations/{name}")




import collections

class object_helper(object):
    """ Helper class for returning objects for nested dictionaries in API Response. """
    def __init__(self, d):
        self.d = d
        for a, b in d.items():
            a = a.replace('-', '_')
            if isinstance(b, (list, tuple)):
                setattr(self, a, [object_helper(x) if isinstance(x, dict) else x for x in b])
            else:
                setattr(self, a, object_helper(b) if isinstance(b, dict) else b)

    def __str__(self):
        if self.d:
            if isinstance(self.d, collections.abc.Iterable):
                return str(self.d)
